﻿EnableExplicit

Structure ProcedureInfo
    include.i
    Map ProcedureReferences.i()
EndStructure

Procedure MarkAsIncluded(Map _root.ProcedureInfo(), Map _referencedProcedures.i())
    Protected name$
    ForEach _referencedProcedures()
        name$ = MapKey( _referencedProcedures() )
        If _root( name$ )\include = 0
            _root( name$ )\include = 1
            MarkAsIncluded(_root(),_root(name$)\ProcedureReferences())
        EndIf
    Next
EndProcedure

Procedure ProcessASM(inputFileName.s, Map _out.ProcedureInfo())
    Protected inputFile
    Protected currentLine$, trimmedLine$, currentLineNum
    Protected inProcedure, currentProcedureName$, regExResultCount, i
    
    Protected regExLabels, regExPBLabels, regExPBLabelsRef, regExProcedureLabels, regExProcedureRef, regExProcedureMacroUsed
    Protected Dim regExResult.s(0)
    Protected procedureCount
    
    ClearMap( _out() )
    
    regExLabels             = CreateRegularExpression(#PB_Any,"^( )*([a-z]|[A-Z]|[0-9]|_)*(\056)*([a-z]|[A-Z]|[0-9]|_)+:",#PB_RegularExpression_NoCase)    ; GENERAL LABELS
    regExPBLabels           = CreateRegularExpression(#PB_Any,"^( )*([a-z]|[A-Z]|[0-9]|_)*(\056)*l_([a-z]|[A-Z]|[0-9]|_)+:",#PB_RegularExpression_NoCase)  ; PB labels, starting with 'l_'
    regExPBLabelsRef        = CreateRegularExpression(#PB_Any,"([a-z]|[A-Z]|[0-9]|_)*(\056)*l_([a-z]|[A-Z]|[0-9]|_)+",#PB_RegularExpression_NoCase)   ; PB label reference, using 'l_'
    regExProcedureLabels    = CreateRegularExpression(#PB_Any,"^( )*_Procedure([0-9])+:.*",#PB_RegularExpression_NoCase)        ; PB procedure labels, starting with '_Procedure'
    regExProcedureRef       = CreateRegularExpression(#PB_Any,"_Procedure([0-9])+",#PB_RegularExpression_NoCase)                ; PB procedure reference, using '_ProcedureXXX' labels
    ;regExProcedureMacroUsed = CreateRegularExpression(#PB_Any,"^( )*MP([0-9])+.*",#PB_RegularExpression_NoCase)                 ; PB procedure macro used, starting with 'MP'
    
    inputFile = ReadFile(#PB_Any,inputFileName)
    If inputFile = 0
        PrintN("File '"+inputFileName+"' not found.")
        ProcedureReturn
    EndIf
    
    currentProcedureName$ = "1"
    
    Protected stringFormat = ReadStringFormat(inputFile)
    
    While Not Eof(inputFile)
        currentLineNum + 1
        currentLine$ = ReadString(inputFile,stringFormat)
        trimmedLine$ = Trim(currentLine$)
        
        If Asc(trimmedLine$)=';'
            Continue
        ElseIf Left(trimmedLine$,8) = "macro MP"
            ;Debug "(line: "+Str(currentLineNum)+") Found procedure macro: "+Right(trimmedLine$,Len(trimmedLine$)-6)
            procedureCount + 1
            inProcedure = #True
        ElseIf inProcedure = #True And Left(trimmedLine$,1) = "}"
            ;Debug "(line: "+Str(currentLineNum)+") End procedure macro"
            ;Debug ""
            currentProcedureName$ = "1"
            inProcedure = #False
        Else
            If inProcedure = #True And MatchRegularExpression(regExProcedureLabels,trimmedLine$)
                regExResultCount = ExtractRegularExpression(regExProcedureLabels,trimmedLine$,regExResult())
                If regExResultCount
                    regExResultCount - 1
                    For i = 0 To regExResultCount
                        ;Debug "   (line: "+Str(currentLineNum)+") Found PB procedure label inside procedure: "+regExResult(i)
                        currentProcedureName$ = Trim(Trim(regExResult(i),":"))
                        _out(currentProcedureName$)\include = 0
                        ;Debug "   current Procedure is: "+currentProcedureName$
                    Next
                EndIf
            ElseIf MatchRegularExpression(regExProcedureRef,trimmedLine$)
                regExResultCount = ExtractRegularExpression(regExProcedureRef,trimmedLine$,regExResult())
                If regExResultCount
                    regExResultCount - 1
                    For i = 0 To regExResultCount
                        ;Debug "   (line: "+Str(currentLineNum)+") Found reference to procedure: "+regExResult(i)
                        _out(currentProcedureName$)\ProcedureReferences(regExResult(i))=1
                    Next
                EndIf
            EndIf
        EndIf
    Wend
    
    Protected name$
    
    _out("1")\include = 1
    ForEach _out("1")\ProcedureReferences()
        MarkAsIncluded(_out(),_out("1")\ProcedureReferences())
    Next
    
    CloseFile(inputFile)
    
    PrintN("Found "+Str(procedureCount)+" procedure macros.")    
EndProcedure

Procedure ReWriteASM(inputFileName.s, Map _procInfo.ProcedureInfo())
    Protected inputFile, outputFile, currentLineNum, stringFormat
    Protected currentLine$, trimmedLine$
    Protected regExResultCount, i
    Protected Dim regExResult.s(0)
    Protected mpfound, mpremoved
    
    Protected NewList sourceLine.s()
    
    Protected regExProcedureMacroUsed = CreateRegularExpression(#PB_Any,"^( )*MP([0-9])+",#PB_RegularExpression_NoCase) ; PB procedure macro used, starting with 'MP'

    inputFile = ReadFile(#PB_Any,inputFileName)
    If inputFile = 0
        PrintN("File '"+inputFileName+"' not found.")
        ProcedureReturn
    EndIf
    
    stringFormat = ReadStringFormat(inputFile)
    
    While Not Eof(inputFile)
        If AddElement( sourceLine() )
            sourceLine() = ReadString(inputFile,stringFormat)
        EndIf
    Wend
    CloseFile(inputFile)
    
    outputFile = CreateFile(#PB_Any,inputFileName)
    If outputFile = 0
        Debug "Can't create output file: '"+inputFileName+"'."
        ProcedureReturn
    EndIf
    
    
    WriteStringFormat(outputFile,stringFormat)
    
    ForEach sourceLine()
        currentLineNum + 1
        currentLine$ = sourceLine()
        trimmedLine$ = Trim(currentLine$)
        If MatchRegularExpression(regExProcedureMacroUsed,trimmedLine$)
            regExResultCount = ExtractRegularExpression(regExProcedureMacroUsed,trimmedLine$,regExResult())
            If regExResultCount
                regExResultCount - 1
                For i = 0 To regExResultCount
                    mpfound + 1
                    If _procInfo("_Procedure"+Mid(regExResult(i),3))\include = 0
                        mpremoved + 1
                        Debug "(line: "+Str(currentLineNum)+") "+regExResult(i)+ " found, but not required. REMOVING"
                        WriteStringN(outputFile,"; "+trimmedLine$+ "; AUTOMATICALLY REMOVED",stringFormat)
                    Else
                        Debug "(line: "+Str(currentLineNum)+") "+regExResult(i)+ " found. OK."
                        WriteStringN(outputFile,currentLine$,stringFormat)
                    EndIf
                Next
            EndIf
        Else
            WriteStringN(outputFile,currentLine$,stringFormat)
        EndIf
    Next
    
    CloseFile(outputFile)
    
    PrintN("Removed "+Str(mpremoved)+" out of "+Str(mpfound)+" used procedure macros.")
EndProcedure

Procedure ShowProcInfo(Map _ProcInfo.ProcedureInfo())
    Protected name$

    ForEach _ProcInfo()
        name$ = MapKey(_ProcInfo())
        PrintN(name$ + " -> include: " + Str( _ProcInfo()\include ))
        ForEach _ProcInfo()\ProcedureReferences()
            PrintN("   "+MapKey( _ProcInfo()\ProcedureReferences() ))
        Next
    Next
EndProcedure

If OpenConsole()
    PrintN("DeadProcedureRemover uses:")
    PrintN("PCRE regular expression library, written by Philip Hazel, Copyright (c) 1997-2007 University of Cambridge")
    PrintN("For PCRE license see pcre_license.txt")
    PrintN("")
    
    
    NewMap ProcInfo.ProcedureInfo()
    
    Define inputFile.s = "Purebasic.asm"
    
    ProcessASM(inputFile,ProcInfo())
    
    ;ShowProcInfo( ProcInfo() )
    
    ReWriteASM(inputFile,ProcInfo())
EndIf

; IDE Options = PureBasic 5.21 LTS (Windows - x86)
; ExecutableFormat = Console
; CursorPosition = 128
; FirstLine = 102
; Folding = -
; EnableUnicode
; Executable = DeadProcedureRemover.exe
; CPU = 5
; CompileSourceDirectory